(*---------------------------------------------------------------------------*
  INTERFACE  iom_octet_stream.mli

  Copyright (c) 2006, James H. Woodyatt
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Core stream commutation for pipes and stream sockets. *)

(** {6 Overview}

    Pipes and stream sockets are system services that communicate in streams
    of octets, usually with orderly close.  Pipes are simplex communication
    channels and sockets are full duplex.  The classes and functions defined
    here are extensions of the basis defined in {!Iom_stream} for handling
    such octet streams.
*)

(** An exception to indicate that a stream transform requires more input before
    output will be produced.
*)
exception Buffering

(** Use [new fragment more message] to create an event object comprising a
    fragment of the octet stream and an indication of whether the end of the
    octet stream has been reached yet.
*)
class fragment: Iom_stream.more -> Cf_message.t -> object
    inherit Iom_stream.event
    
    (** Use [fragment#data] to obtain a message containing the current fragment
        of the octet stream.
    *)
    method data: Cf_message.t
end

(** A record type used for specifying the low and high limit marks for octet
    stream transforms.  In general, octet stream transforms send [`Wait] when
    they have buffered more than [high] octets, and send [`Ready] when enough
    octets are drained that less than [low] octets remain.
*)
type limits = { low: int; high: int }

(** Use [normalize_limits lim] to obtain a new limits record with [low]
    constrained to be a positive integer, and [high] constrained to be greater
    than or equal to [low].
*)
val normalize_limits: limits -> limits

(** {4 State Transforms } *)

(** Use [inherit \['x, 'y\] transform ~limits flowTx inRx outTx] to define a
    subclass for transforming streams of ['x] events received at [inRx] into
    streams of ['y] events transmitted at [outTx].  Conventionally, both ['x]
    and ['y] are subtypes of [fragment], but this is not required.  Logically,
    events are represented by stream fragments, represented as a blocks of
    octets and an optional sentinel indicating the end of the stream.
*)
class virtual ['x, 'y] transform:
    limits:limits -> [> Iom_stream.readywait ] #Iom_gadget.tx ->
    'x #Iom_gadget.rx -> 'y #Iom_gadget.tx ->
    object
        inherit ['x, 'y] Iom_layer.transform

        (** The pending octet buffer. *)
        val buffer_: Cf_message.substring Cf_deque.t
        
        (** The number of octets in the buffer. *)
        val mark_: int
        
        (** A flag to specify whether the input stream has ended. *)
        val more_: Iom_stream.more
        
        (** The limits record for this transform. *)
        val limits_: limits
        
        (** Evaluate the result of [self#push_octets message] to obtain the new
            [consume_] flag, the new [buffer_] and the new [mark_] after
            pushing the octets in [message].
        *)
        method private push_octets:
            Iom_stream.more -> Cf_message.t ->
            (bool * Cf_message.substring Cf_deque.t * int) Iom_gadget.t
        
        (** Evaluate the result of [self#shift_octets length] to obtain the new
            [consume_] flag, the new [buffer_] and the new [mark_] after
            shifting a maximum of [length] octets from the buffer.
        *)
        method private shift_octets:
            int -> (bool * Cf_message.substring Cf_deque.t * int) Iom_gadget.t
    end

(** Use [inherit emitter ?limits flowTx inRx outTx] to derive an octet stream
    transform class (for use with {!Iom_layer.osimplex} or
    {!Iom_layer.duplex}).  Override the [self#fragment] and [self#emit] methods
    in the subtype to implement the discipline for producing events to output.
*)
class ['x, 'y] emitter:
    ?limits:limits -> [> Iom_stream.readywait ] #Iom_gadget.tx ->
    'x #Iom_gadget.rx -> 'y #Iom_gadget.tx ->
    object('self)
        inherit ['x, 'y] transform
        constraint 'y = #fragment

        (** Override [self#fragment inEvent] to return the octet stream
            fragment components of the input event to the basis implementation
            of [self#push].  Basis method returns [Iom_stream.Last] and an
            empty message fragment.
        *)
        method private fragment: 'x -> Iom_stream.more * Cf_message.t
        
        (** Override [self#emit] to examine the contents of the current buffer,
            identify all the output events that can be shifted, and the length
            of the octets in the buffer to shift.  Basis method return an empty
            list of events and zero length for the shift.
        *)
        method private emit: 'y list * int

        (** This method is invoked by [self#consume] to push the input event.
            It calls [self#fragment] to obtain the [more_] flag and a message
            fragment, which it applies to [self#push_octets] and returns a
            new transform state.
        *)
        method push: 'x -> 'self Iom_gadget.t
        
        (** This method is invoked by [self#produce] to shift zero or more
            output events from the buffer.  It calls [self#emit], then calls
            [self#shift_octets] with the length obtained, and returns a new
            transform state.
        *)
        method shift: ('y list * 'self option) Cf_exnopt.t Iom_gadget.t
    end

(** Use [inherit scanner ?limits flowTx inRx outTx] to derive an octet stream
    transform class (for use with {!Iom_layer.isimplex} or
    {!Iom_layer.duplex}).  Override the [self#lex] and/or [self#parse] methods
    in the subtype to implement the discipline for producing events to output.
*)
class ['x, 'y] scanner:
    ?limits:limits -> [> Iom_stream.readywait ] #Iom_gadget.tx ->
    'x #Iom_gadget.rx -> 'y #Iom_gadget.tx ->
    object('self)
        inherit ['x, 'y] transform
        constraint 'x = #fragment

        (** Override [self#lex message] to provide a functional parser to
            [self#parse] that will be applied to the character sequence in
            [message] to obtain an output event and the number of octets to
            shift.  The basis implementation always returns the [nil] parser.
        *)
        method private lex: Cf_message.t -> ('y * int) Cf_lex.t
        
        (** Parse the octet buffer by calling [self#lex] in a loop until no
            more output objects are scanned.  Returns [None] if no output is
            recognized, otherwise returns [Some (output, length)] where
            [output] is a list of events of type ['y] and [length] is the
            number octets to shift with [self#shift_octets] in the [shift]
            method.
        *)
        method private parse: ('y list * int) option

        (** This method is invoked by [self#consume] to push the input event.
            It decomposes the input fragment into its [more_] flag and its
            octet fragment, which it applies to [self#push_octets] and returns
            a new transform state.
        *)
        method push: 'x -> 'self Iom_gadget.t
        
        (** This method is invoked by [self#produce] to shift zero or more
            output events from the buffer.  It calls [self#parse], and if some
            events were returned, then calls [self#shift_octets] with the
            length obtained, and returns a new transform state.
        *)
        method shift: ('y list * 'self option) Cf_exnopt.t Iom_gadget.t
    end

(** {4 Functions} *)

(** Use [input_buffer ?limits fix] to apply the buffering discipline to the
    input stream in [fix] according to [limits].
*)
val input_buffer:
    ?limits:limits -> (fragment, 'c, 'n) Iom_stream.ifix ->
    (fragment, 'c, 'n) Iom_stream.ifix Iom_gadget.t

(** Use [output_buffer ?limits fix] to apply the buffering discipline to the
    output stream in [fix] according to [limits].
*)
val output_buffer:
    ?limits:limits -> (fragment, 'c, 'n) Iom_stream.ofix ->
    (fragment, 'c, 'n) Iom_stream.ofix Iom_gadget.t

(** Use [duplex_buffers ?ilim ?olim fix] to apply the buffering discipline to
    the output stream in [fix] according to [ilim] and [olim].
*)
val duplex_buffers:
    ?ilim:limits -> ?olim:limits ->
    (fragment, fragment, 'c, 'n) Iom_stream.iofix ->
    (fragment, fragment, 'c, 'n) Iom_stream.iofix Iom_gadget.t

(*--- End of File [ iom_octet_stream.mli ] ---*)
